﻿using FFmpeg.AutoGen.Example;
using OpenCvSharp;
using SharpDX;
using System;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using WindowsFormsApp1.FFmpeg;

namespace WindowsFormsApp1
{
    public partial class Form1 : Form
    {
        EventWaitHandle _exitEvent = new EventWaitHandle(false, EventResetMode.ManualReset);
        EventWaitHandle _captureEvent = new EventWaitHandle(false, EventResetMode.AutoReset);
        EventWaitHandle[] _waitSignals;
        Thread _captureThread;

        CaptureSource _captureSource;

        public Form1()
        {
            InitializeComponent();

            this.KeyPreview = true;

            _captureSource = CaptureSource.Monitor1;
            _waitSignals = new EventWaitHandle[] { _exitEvent, _captureEvent };
            _captureThread = new Thread(captureThreadFunc);
            _captureThread.IsBackground = true;
            _captureThread.Start();

            FFmpegBinariesHelper.RegisterFFmpegBinaries();
        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            _exitEvent.Set();

            base.OnFormClosing(e);
        }

        bool _exitCapture = false;

        protected override void OnKeyPress(KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Cancel) // Ctrl + C
            {
                _exitCapture = false;
                _captureEvent.Set();
            }

            if (e.KeyChar == 0x18) // Ctrl + X
            {
                _exitCapture = true;
            }

            base.OnKeyPress(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
        }

        private void captureThreadFunc()
        {
            DXGIManager manager = null;

            while (true)
            {
                using (manager = new DXGIManager(_captureSource))
                {
                    if (manager.Initialized == false)
                    {
                        return;
                    }

                    // Initialize 호출 후 너무 빠르게 Capture를 호출하면 (검은 색의) 빈 화면이 나올 수 있음.
                    // 필요하다면 Sleep을 호출
                    // Thread.Sleep(500);

                    try
                    {
                        while (true)
                        {
                            int signalId = EventWaitHandle.WaitAny(_waitSignals, Timeout.Infinite);

                            if (signalId == 0)
                            {
                                break;
                            }

                            // Ctrl + C로 시작해 Ctrl + X가 눌릴 때까지 화면을 캡처해 동영상으로 저장
                            CaptureLoopMat(manager);
                        }

                        break;
                    }
                    catch (SharpDXException e)
                    {
                        if (e.ResultCode != DXGIError.DXGI_ERROR_ACCESS_LOST)
                        {
                            // 예를 들어, Ctrl + Shift + Alt 키를 눌러 데스크탑 전환을 한 경우
                            throw;
                        }
                    }
                }
            }
        }

        private void CaptureLoopMat(DXGIManager manager)
        {
            int fps = 30;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            OpenCvSharp.Size frameSize = new OpenCvSharp.Size(manager.Width, manager.Height);
            using (Mat frame = new Mat(frameSize, MatType.CV_8UC4))
            using (FFmpegVideoWriter writer = new FFmpegVideoWriter("C:\\temp\\test.avi", FourCC.XVID, fps, frameSize, true))
            using (FpsWriter matBuffer = new FpsWriter(writer, MatType.CV_8UC3))
            {
                while (_exitCapture == false)
                {
                    long started = sw.ElapsedMilliseconds;

                    if (manager.Capture(frame.Data, 1000) == true)
                    {
                        Mat current = matBuffer.GetMat();
                        Cv2.CvtColor(frame, current, ColorConversionCodes.BGRA2BGR);
                        matBuffer.Write(current);
                    }
                    else
                    {
                        matBuffer.MarkElapsed(1000);
                    }

                    int elapsed = (int)(sw.ElapsedMilliseconds - started);
                    matBuffer.TrySleep(elapsed);
                }
            }
        }
    }
}
